/*
    FEARLUS model0-7-1ci: CAVNT3Clumper.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1ci, an agent-based model of land use
    change.

    FEARLUS model0-7-1ci is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1ci is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation of Cellular Automaton Von-Neumann Toroidal clumper that changes
the value of a bit in the land parcel bit string if 3 or more neighbours have
a differently valued bit.

*/

#import "CAVNT3Clumper.h"
#import "Environment.h"
#import "LandParcel.h"
#import "FearlusArguments.h"
#import "MiscFunc.h"
#import "Parameter.h"
#import "BitString.h"
#import <string.h>

@implementation CAVNT3Clumper

/*

setParameters:

Pass the parameter object into the clumper

*/

-(void)setParameters: (Parameter *)p {
  parameter = p;
}

/*

setParameter:toValue:

This is a means of passing various clump-class specific parameters in to the
clump object.

*/

-(BOOL)setParameter: (char *)param toValue: (char *)value {
  if(strcmp(param, "Cycles") == 0) {
    clumpCycles = (unsigned)atoi(value);
    return YES;
  }
  else {
    fprintf(stderr, "CAVNT3Clumper does not have parameter %s\n", param);
    return NO;
  }
}

/*

clumpEnvironment:

Use a CA-type algorithm to clump the land parcel characteristics together. 
Each bit in the land parcel bit string is taken as a characteristic of the
land parcel. For a specified number of times (clumpCycles), the following is
done for each characteristic. (Assuming initial random distribution of
characteristic values.)

1.	For each cell, compute a bit string as follows:
	If three or more neighbours differ in value for that bit then 1 else 0
2.	Toggle any bit with 1 in the corresponding place in the stored string.

(The toggle operation is functionally equivalent to XOR.)

*/

-(void)clumpEnvironment: (Environment *)env {
  unsigned clumpCycle, attr;
  int x, y, bit;
  id toggleMatrix, tempZone, aNeighbour;
  BitString *bs;

  if([FearlusArguments debugCoreMethods]) {
    printf("E: %s: Clumping (CAVNT3Clumper) (Attributes displayed each cycle)"
	   "\n", sel_get_name(_cmd));
  }
  else {
    printf("Clumping (CAVNT3Clumper) (1 dot for each cycle) ");
  }
  fflush(stdout);

  /* Create a grid of bit strings, one for each item in the grid. Each bit
     string contains a 1 if the corresponding bit in the corresponding grid
     cell in the environment is to be changed. */
  tempZone = [Zone create: scratchZone];
  toggleMatrix = [Discrete2d createBegin: tempZone];
  [toggleMatrix setSizeX: [env getSizeX] Y: [env getSizeY]];
  toggleMatrix = [toggleMatrix createEnd];
  for(x = 0; x < [toggleMatrix getSizeX]; x++) {
    for(y = 0; y < [toggleMatrix getSizeY]; y++) {
      bs = [BitString create: tempZone
		      setBitCount: [parameter parcelBitStringSize]];
      [toggleMatrix putObject: bs atX: x Y: y];
    }
  }

  /* Loop through the clump cycles */
  for(clumpCycle = 0; clumpCycle < clumpCycles; clumpCycle++) {

    /* Create the toggle bit string for each cell in the environment */
    for(x = 0; x < [env getSizeX]; x++) {
      for(y = 0; y < [env getSizeY]; y++) {
	bs = [toggleMatrix getObjectAtX: x Y: y];
	for(attr = 0; attr < [parameter parcelBitStringSize]; attr++) {
	  int nx, ny, c;
	  bit = 0;			// Sum of bits that are different

	  /* This loop cycles through the von Neumann (toroidal) neighbours of
	     the current cell. nx and ny cycle through the following values as
	     c increases:

	     c	 0  1  2  3
	     nx -1  0  1  0
	     ny  0  1  0 -1

	  */
	  for(nx = -1, ny = 0, c = 0;
	      c < 4;
	      c++, nx = ny, ny = ((c % 2) - (2 * (c / 2))) * (c % 2)) {
	    int nxc, nyc;

	    /* nxc and nyc are the actual co-ordinates in the grid of the
	       neighbouring cell */
	    nxc = [MiscFunc Mod: (x + nx) inRange0To: [env getSizeX]];
	    nyc = [MiscFunc Mod: (y + ny) inRange0To: [env getSizeY]];

	    aNeighbour = [env getObjectAtX: nxc Y: nyc];
	    if([aNeighbour getOneAttribute: attr]
	       != [[env getObjectAtX: x Y: y] getOneAttribute: attr]) {
	      bit++;			// Increment bits that are different
	    }
	  }
	  if(bit >= CLUMPTHRESHOLD) {
	    [bs setBit: attr to: 1];	// Set this bit to be toggled
	  }
	  else {
	    [bs setBit: attr to: 0];
	  }
	}
      }
    }
  
    /* Toggle all the bitstrings */

    for(x = 0; x < [env getSizeX]; x++) {
      for(y = 0; y < [env getSizeY]; y++) {
	[[env getObjectAtX: x Y: y] toggleBits:
				     [toggleMatrix getObjectAtX: x Y: y]];
      }
    }

    /* If requested, print out information on the results of clumping */
    if([FearlusArguments debugCoreMethods]) {
      printf("E: %s: Results of clumping after %u cycles:\n",
	     sel_get_name(_cmd), clumpCycle + 1);
      [env printTextParcelBitstrings];
    }
    else {
      printf(".");
    }
    fflush(stdout);
  }
  printf("\n");
  [tempZone drop];
}

@end
