/*
    FEARLUS model0-7-1ci: ImitativeOptimumStrategyRational.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1ci, an agent-based model of land use
    change.

    FEARLUS model0-7-1ci is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1ci is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the ImitativeOptimumStrategyRational object. 

*/

#import "ImitativeOptimumStrategyRational.h"
#import "SelectUseBucketRationals.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "LandManager.h"
#import "Environment.h"
#import "Parameter.h"
#import "FearlusArguments.h"
#import "BitString.h"
#import "MiscFunc.h"
#import "Interval.h"
#import <DoubleRationalWarn.h>

@implementation ImitativeOptimumStrategyRational
/*

create:withManager:andParameters:

Create the strategy, setting the manager and parameters.

*/

+(id <ImitativeStrategy, HistoricalStrategy>)create: aZone
	   withManager: (LandManager *)lmgr
	 andParameters: (Parameter *)p {
  ImitativeOptimumStrategyRational *obj;

  obj = [super create: aZone];
  obj->lm = lmgr;
  obj->parameter = p;
  return obj;
}

/*

decideLandUseForParcel:

Decide the land use for the land parcel, using an imitative optimum
strategy.  A formula is used to calculate the expected yield for the
decision parcel given the yield, land use and bit string of the
neighbouring parcel. A single exemplar is sought for each land use, if
it appears in the neighbourhood, as theoretically the expected yield
should always be the same. This expected yield is stored as a rational
number.

The buckets will then contain the expected yield obtained for each
land use observed for the decision parcel. A random selection is made
among those with the highest score. The expected yield calculation
will not cope with don't care bits (when present, these land managers
will no longer be imitative *optimisers*).

This method does make use of neighbourhood weighting and neighbourhood
noise.  Neighbourhood noise will disrupt the bit string of the
neighbouring parcels belonging to other land managers. The effect will
be some miscalculation of the expected yield. Neighbourhood weighting
can be used to mitigate these effects.  A lower neighbourhood
weighting will mean land managers are less likely to use land uses
that they do not already use.

The land manager's own land parcels are always inspected first. Thus
any neighbour's parcel with the same land use will be ignored. This
will further mitigate the effects of neighbourhood noise.

*/

-(LandUse *)decideLandUseForParcel: (LandParcel *)lp {
  SelectUseBucketRationals *bucket;
  id <Index> parcelIndex;
  id tempZone;
  LandParcel *myParcel, *neighbouringParcel;
  LandManager *neighbour;
  LandUse *lu;
  int i;
  unsigned yield;
  unsigned match;
  BitString *bs;
  DoubleRationalWarn *getExpectedYield(unsigned yield, unsigned antimatch,
			     unsigned lubits) {
    int i;
    DoubleRationalWarn *eyieldRational = 
      [DoubleRationalWarn parseKeepSameClass: "DoubleRationalWarn=0!ALL"];
    DoubleRationalWarn *tmpRat = [DoubleRationalWarn new];
    
    [eyieldRational cset: 0.0];
    if([FearlusArguments debugInnerCoreMethods]) {
      printf("i: Calculation of Expected Yield:\n\tYield = %u, antimatch = %u,"
	     " land use bits = %u\n", yield, antimatch, lubits);
      printf("\tYield        No. of ways             Out of   Probability\n");
    }
    for(i = 0; i <= (int)yield; i++) {

      [tmpRat cset: (double)((int)yield + (int)antimatch - (2 * i))];
      [tmpRat cmul: (double)[MiscFunc combineN: (int)antimatch r: i]];
      [tmpRat cmul: (double)[MiscFunc combineN: (int)(lubits - antimatch)
					   r: (int)yield - i]];
      [eyieldRational add: tmpRat];
      
      if([FearlusArguments debugInnerCoreMethods]) {
	unsigned long long ways, total;

	ways = [MiscFunc combineN: (int)antimatch r: i]
	  * [MiscFunc combineN: (int)(lubits - antimatch) r: (int)yield - i];
	total = [MiscFunc combineN: (int)lubits r: (int)yield];
	printf("\t% 5d % 18lld % 18lld % 13g\n",
	       ((int)yield + (int)antimatch - (2 * i)), ways, total,
	       (total > 0 ? (double)ways / (double)total : 0.0));
      }
    }
    [tmpRat free];

    [eyieldRational cdiv: 
		      (double)[MiscFunc combineN: (int)lubits r: (int)yield]];
    
    return eyieldRational;
  }

  tempZone = [Zone create: scratchZone];
  bucket = [SelectUseBucketRationals create: tempZone
			     withParameters: parameter
			     andLandUses: [[lp getEnvironment] getLandUses]];
  // Loop through land parcels owned by the land manager
  if([FearlusArguments debugCoreMethods]) {
    printf("S: %s: Using imitative optimum strategy of land manager %u to "
	   "determine land use for LP = land parcel %u at (%d, %d)\n",
	   sel_get_name(_cmd), [lm getPIN], [lp getPIN], [lp getX], [lp getY]);
  }
  for(parcelIndex = [[lm getLandParcels] begin: tempZone], [parcelIndex next];
      [parcelIndex getLoc] == Member;
      [parcelIndex next]) {
    DoubleRationalWarn *tmpRational;
    myParcel = [parcelIndex get];
    yield = [myParcel getLastYield];
    match = [[lp getAttributes] getMatchWith: [myParcel getAttributes]];

    tmpRational = [bucket getScoreInBucketOfLandUse: [myParcel getLandUse]];
    if([tmpRational ceq: 0.0]) {
      DoubleRationalWarn *score;

      score = getExpectedYield(yield,
			       [parameter parcelBitStringSize] - match,
			       [parameter useBitStringSize]);
      if([FearlusArguments debugCoreMethods]) {
	printf("S: %s: Found new land use %u at land parcel %u at (%d, %d) "
	       " with antimatch to LP of %u and yield %u. Expected yield "
	       " (= score for LU) = ", sel_get_name(_cmd),
	       [[myParcel getLandUse] getPIN], [myParcel getPIN],
	       [myParcel getX], [myParcel getY], 
	       [parameter parcelBitStringSize] - match, yield);
	[score write: stdout];
	printf("\n");
      }
      [bucket addScore: score toBucketForLandUse: [myParcel getLandUse]];
      [myParcel incNImitations];
    }
    else {
      if([FearlusArguments debugCoreMethods]) {
	DoubleRationalWarn *anotherTmp;

	printf("S: %s: Already got expected yield (", sel_get_name(_cmd));
	[tmpRational write:stdout];
	printf(") for land use %u of land parcel %u at (%d, %d)."
	       " Recalculation of expected yield gives ",
	       [[myParcel getLandUse] getPIN], [myParcel getPIN],
	       [myParcel getX], [myParcel getY]); 
	
	anotherTmp = getExpectedYield(yield,
				      [parameter parcelBitStringSize] - match,
				      [parameter useBitStringSize]);
	[anotherTmp write: stdout];
	[anotherTmp free];
	printf("\n");
      }
    }
    [tmpRational free];
  }
  [parcelIndex drop];
  // Loop through the neighbours, and through their land parcels
  bs = [BitString create: tempZone
		  setBitCount: [parameter parcelBitStringSize]];
  for(i = 0; (neighbour = [lm getNextNeighbour: i]); i |= 1) {
    if([FearlusArguments debugCoreMethods]) {
      printf("S: %s: Adding scores for land uses of land parcels belonging"
	     " to land manager %u\n", sel_get_name(_cmd), [neighbour getPIN]);
    }
    for(parcelIndex = [[neighbour getLandParcels] begin: tempZone],
	  [parcelIndex next];
	[parcelIndex getLoc] == Member;
	[parcelIndex next]) {
      DoubleRationalWarn *tmpRational;
      neighbouringParcel = [parcelIndex get];
      [bs setToCopyOf: [neighbouringParcel getAttributes]];
      [bs mutateWithProbability: [lm getNeighbourNoise]];
      match = [[lp getAttributes] getMatchWith: bs];
      yield = [neighbouringParcel getLastYield];

      tmpRational = [bucket getScoreInBucketOfLandUse: 
			      [neighbouringParcel getLandUse]];
   
      if([tmpRational ceq: 0.0]) {
	DoubleRationalWarn *score;

	score = getExpectedYield(yield,
				 [parameter parcelBitStringSize] - match,
				 [parameter useBitStringSize]);

	/* Now we should multiply score by [lm getNeighbourWeight], but 
	   [lm getNeighbourWeight] is an interval, so we have a problem. 
	   Solution: abort if [lm getNeighbourWeight] != 0.0 */
	if ([[lm getNeighbourWeight] ne: [Interval constant: 1.0]]) {
	  fprintf(stderr, "%s -- ImitativeOptimumStrategyRational does not "
		  "work with neighbouring weighting\n", sel_get_name(_cmd));
	  exit(1);
	}

	if([FearlusArguments debugCoreMethods]) {
	  printf("S: %s: Found new land use %u at neighbour's land parcel %u "
		 "at (%d, %d) with antimatch to LP of %u and yield %u. "
		 "Expected yield = ", sel_get_name(_cmd),
		 [[neighbouringParcel getLandUse] getPIN], 
		 [neighbouringParcel getPIN], [neighbouringParcel getX],
		 [neighbouringParcel getY],
		 [parameter parcelBitStringSize] - match, yield);
	  [score write: stdout];
	  printf(". Score added = ");
	  [score write: stdout];
	  printf("\n");
	}
	[bucket addScore: score 
		toBucketForLandUse: [neighbouringParcel getLandUse]];
	[neighbouringParcel incNImitations];

      }
      else {
	if([FearlusArguments debugCoreMethods]) {

	  DoubleRationalWarn *anotherTmp;

	  printf("S: %s: Already got expected yield (", sel_get_name(_cmd));
	  [tmpRational write:stdout];
	  printf(") for land use %u of land parcel %u at (%d, %d)."
	       " Recalculation of expected yield gives ",
		 [[neighbouringParcel getLandUse] getPIN],
		 [neighbouringParcel getPIN],
		 [neighbouringParcel getX], [neighbouringParcel getY]);
	
	  anotherTmp = getExpectedYield(yield,
				      [parameter parcelBitStringSize] - match,
				      [parameter useBitStringSize]);
	  [anotherTmp write: stdout];
	  [anotherTmp free];
	  printf("\n");
	}
      }
      [tmpRational free];
    }
    [parcelIndex drop];
  }
  lu = [bucket getRandomOptimumChoice];
  [bucket drop];
  [bs drop];
  [tempZone drop];
  return lu;
}

@end
