/*
    FEARLUS model0-7-1ci: LandManager.h
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1ci, an agent-based model of land use
    change.

    FEARLUS model0-7-1ci is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1ci is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Interface for the LandManager class. Land Managers have a list of land
parcels they manage, and a list of neighbouring land managers. A
neighbour of a land manager is defined as the list of managers who
neighbour any parcel owned.  They also have a land use strategy. This
is implemented as a floating point weighting indicating the degree of
influence a land manager's neighbours have on their selection of a
land use. There is a history of land use strategies.

0-5-3: Changed yieldThreshold from unsigned to double

NOTE: There are a number of variables and methods with "barn" in their
name -- a legacy of an earlier version of FEARLUS. These should all be
thought of has referring to the wealth of the land manager.

NOTE: Another legacy from earlier versions is the use of "strategy" to
refer to "neighbourhood weighting". This will be flagged in the
comments when it applies.

NOTE: A further legacy from earlier versions is the use of "habit
threshold" or "habit imitation threshold" to refer to what we now call
"aspiration threshold". Various method and variable names referring to
"habit" should be interpreted as referring to "aspiration".

 */

#import <stdlib.h>
#import <stdio.h>
#import <errno.h>
#import <objectbase/SwarmObject.h>
#import <random.h>
#import <collections.h>			// ListShuffler
#import <objc/objc-api.h>		// sel_get_name
#import "Strategy.h"

@class Environment, LandAllocator, LandParcel, Parameter, SubPopulation,
  BitString, Tube, Interval;


@interface LandManager: SwarmObject {
  // Instance variables with a causal influence
  Parameter *parameter;
  SubPopulation *subPop;	// Which sub-population this manager belongs to
  Interval *neighbourWeight;	// Amount of attention paid to neighbours
  Interval *pImitative;		// Probability of using an imitative strategy
  Interval *nbrWgtChangeUnits;	// A model parameter which tells the
				// land manager how much to update
				// their neighbourhood weighting by
  Interval *neighbourMutateProb;
				// A model parameter representing incomplete
				// knowledge about neighbours' land parcels
  id <List> landParcelsOwned;
  Environment *environment;
  LandAllocator *landAllocator;
  Interval *aspirationThreshold;
				/* The aspiration threshold of the land manager
				   -- how much yield they hope to get from
				   each land parcel */
  Interval *wealth;		/* Land managers can store their surplus yield
				   in model 0-3. This is the variable to do it
				   with.*/
  unsigned memorySize;		/* The size of the memory for this land manager
				   The strategy object can look back this far
				   into the past. */
  Tube *climateMemory;		/* The land manager's memory of the climate */
  Tube *economyMemory;		/* The land manager's memory of the economy */
  /* Since it makes sense for the fixed land parcel price and the break even
     threshold to both be floating point numbers, then the wealth must be a
     floating point number too. */
  id <NonHistoricalStrategy, NonImitativeStrategy> initialStrategy;
  id <Strategy> atOrAboveThresholdStrategy;
  id <ImitativeStrategy> belowThresholdImitativeStrategy;
  id <NonImitativeStrategy> belowThresholdNonImitativeStrategy;
  // Instance variables used for internal machinations
  BOOL someLandParcelsLost;
  BOOL allLandParcelsLost;
  id <List> neighbourList;
  id <Index> index;		// This is used in the getNextNeighbour method
  int lpx, lpy;
  int numberOfLandParcelsGainedThisYear;
  // Instance variables for observation purposes
  id <List> history;
  int colour;
  unsigned pin;
  int alive;
  unsigned age;
}

+create: (id)z;
-(unsigned)getPIN;
-(void)setParameters: (Parameter *)p;
-(void)setSubPopulation: (SubPopulation *)sp;
-(SubPopulation *)getSubPopulation;
-(void)kill;
-(void)initialiseWithEnvironment: (Environment *)e
		   landAllocator: (LandAllocator *)la
			  colour: (int)col;
-(LandManager *)getNextNeighbour: (int)i;
-(void)getSocialNeighbourList: (id <List>)l;
-(void)getPhysicalNeighbourList: (id <List>)l;
-(void)allocateLandUses;
-(void)allocateInitialLandUses;
-(void)harvest;
-(void)sellLandParcels;
-getLandParcels;
-(Interval *)getWealth;
-(double)getWealthForDisplayOnly;
				/* used by observer */
-(int)getAgeAsInt;
-(void)incAge;
-(unsigned)getMemorySize;
-(BitString *)recallClimateAgo: (unsigned)yrs;
-(BitString *)recallEconomyAgo: (unsigned)yrs;
-(unsigned)recallYieldForParcel: (LandParcel *)lp yearsAgo: (unsigned)yrs;
-(LandUse *)recallLandUseForParcel: (LandParcel *)lp yearsAgo: (unsigned)yrs;
-(int)getNumberOfLandParcelsOwned;
-(void)addLandParcel: (LandParcel *)lp;
-(void)updateNeighbourhoodWeighting;
-(void)recordHistory;
-(BOOL)eligibleForNewLandParcels;
-(BOOL)eligibleForRemovalFromGame;
-(void)printHistoryToFile: (FILE *)fp;
-(int)getColour;
-(Interval *)getAspiration;
-(double)getAspirationThresholdForDisplayOnly;
                                // Used by observer
-(Interval *)getImitateProb;
-(double)getImitateProbForDisplayOnly;
				// Used by observer
-(Interval *)getNeighbourWeight;
-(Interval *)getNeighbourNoise;
-(void)drop;

@end
