/*
    FEARLUS model0-7-1ci: LandManager.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1ci, an agent-based model of land use
    change.

    FEARLUS model0-7-1ci is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1ci is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the LandManager class.

*/

#import "LandManager.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "Environment.h"
#import "BitString.h"
#import "LandManagerHistory.h"
#import "LandAllocator.h"
#import "FearlusArguments.h"
#import "Parameter.h"
#import "SubPopulation.h"
#import "MiscFunc.h"
#import "Tube.h"
#import "Interval.h"

@implementation LandManager

/*

create

Override the superclass create method to implement a unique identifier system.
This uses a static local variable in the create method. Static variables in
methods are shared across all instances of a class.

*/

+create: (id)z {
  LandManager *new;
  static unsigned counter = 1;

  new = [super create: z];
  new->pin = counter;
  new->alive = 1;
  new->subPop = nil;
  new->wealth = [Interval createWith: 0.0];
  new->neighbourWeight = [Interval create];
  new->pImitative = [Interval create];
  new->nbrWgtChangeUnits = [Interval create];
  new->neighbourMutateProb = [Interval create];
  new->aspirationThreshold = [Interval create];
  new->memorySize = 1;
  counter++;
  return(new);
}

/*

getPIN

Return the unique identification number of this instance.

*/

-(unsigned)getPIN {
  return(pin);
}

/*

setParameters:

Pass in the model parameters to the object

*/

-(void)setParameters: (Parameter *)p {
  parameter = p;
}

/*

setSubPopulation:

Set which sub-population the land manager belongs to. This method must be
called before initialise...

*/

-(void)setSubPopulation: (SubPopulation *)sp {
  subPop = sp;
}

/*

getSubPopulation

Return the sub-population the land manager belongs to.

*/

-(SubPopulation *)getSubPopulation {
  return subPop;
}

/*

kill

Tell the land manager they are out of the game.

*/

-(void)kill {
  alive = 0;
}

/*

initialiseWithEnvironment

After creating, pass in parameters to initialise the object. Also
initialise the history and neighbours lists, and the list of land
parcels owned.  Initialise the decision making algorithm (strategy
parameters and types).

*/

-(void)initialiseWithEnvironment: (Environment *)e
		   landAllocator: (LandAllocator *)la
			  colour: (int)col {
  if(subPop == nil) {
    fprintf(stderr, "%s -- Land Manager must have sub-population set before "
	    "initialisation\n", sel_get_name(_cmd));
    exit(1);
  }
  environment = e;
  landAllocator = la;
  [aspirationThreshold set: [subPop getAHabitImitateThreshold]];
  [neighbourWeight set: [subPop getANeighbourWeight]];
  if([neighbourWeight lt: [Interval constant: 0.0]]) {
    [neighbourWeight set: [Interval constant: 0.0]];
  }
  [pImitative set: [subPop getAnImitateProb]];
  memorySize = [subPop getAMemorySize];
  climateMemory = [Tube create: [self getZone] setLength: memorySize];
  economyMemory = [Tube create: [self getZone] setLength: memorySize];
  atOrAboveThresholdStrategy = [subPop getAnAboveStrategyForLandManager: self
				       andParameters: parameter];
  belowThresholdNonImitativeStrategy
    = [subPop getABelowNonImitativeStrategyForLandManager: self
	      andParameters: parameter];
  belowThresholdImitativeStrategy
    = [subPop getABelowImitativeStrategyForLandManager: self
	      andParameters: parameter];
  initialStrategy = [subPop getAnInitialStrategyForLandManager: self
			    andParameters: parameter];
  [neighbourMutateProb
    set:
      [Interval constant: [uniformDblRand
			    getDoubleWithMin: [parameter neighbourNoiseMin]
			    withMax: [parameter neighbourNoiseMax]]]] ;
  [nbrWgtChangeUnits set: [parameter strategyChangeUnit]];
  neighbourList = [List create: [self getZone]];
  history = [List create: [self getZone]];
  landParcelsOwned = [List create: [self getZone]];
  [wealth set: [Interval constant: 0.0]];
  
  age = 0;
  index = nil;
  colour = col;
}

/*

getNextNeighbour

Return the next land manager that is a neighbour of self. The integer argument
is a flag. When 0, it will restart the list. Any other value, and it returns
the next one. The method works by interrogating each neighbour of the land
parcels owned by self to get the land manager of that land parcel. It must keep
a list of the neighbouring land managers found so far, in order to avoid
duplicates. A land manager might also happen to own neighbouring land parcels
themselves, and in this case they should not be returned as being their own
neighbour. Return nil when there are no more neighbours.

Here's an example loop:

for(i = 0; neighbour = [landManager getNextNeighbour: i]; i |= 1) {

  // Do your stuff

}

Use i |= 1 rather than i++, because if the system has 2 byte integers, and
the world is large enough, and a manager is successful enough, there might be
enough neighbours to go back to 0. OK, so that would require a land manager to
have 65536 or more neighbours -- which, with von Neumann neighbourhood requires
at least 16384 land parcels to be owned, but better safe than sorry.

NB1 (Very Important): It is imperative that this method be used in a loop, with
no other processing going on. The method uses the environment's get next
neighbour method, and it can only be used consecutively for one object at a
time. So, for example, if I ask for the next neighbour of X, it will start
processing X, but if I ask for the next neighbour of Y when I haven't finished
with X, it will start processing Y, and when I go back to X, it will start X
at the beginning again.

NB2: Because of NB1, this method is going to become deprecated, and the
getSocialNeighbourList: method should be used in its place.

*/

-(LandManager *)getNextNeighbour: (int)i {
  LandManager *nlm;

  if(i == 0) {			// Restart finding neighbouring land managers
    if(index != nil) [index drop];
    [neighbourList removeAll];	// Clear the neighbours list.
    [self getSocialNeighbourList: neighbourList];
    [MiscFunc shuffleList: neighbourList];
				// Make sure neighbours are inspected in a
				// random order
    index = [neighbourList begin: scratchZone];
    [index next];		// Point the index to the first neighbour
  }
  // Return the neighbour on the list, or nil if we're at the end
  if([index getLoc] != Member) {
    return nil;
  }
  else {
    nlm = (LandManager *)[index get];
    [index next];
    return nlm;
  }
}

/*

getSocialNeighbourList:

Returns a list of land managers owning land parcels that neighbour the
land manager. The argument is a pre-allocated list pointer.

*/

-(void)getSocialNeighbourList: (id <List>)l {
  id <Index> nlpi;
  id <List> nlpl;
  LandParcel *nlp;

  nlpl = [List create: scratchZone];
  [self getPhysicalNeighbourList: nlpl];
  for(nlpi = [nlpl begin: scratchZone], nlp = (LandParcel *)[nlpi next];
      [nlpi getLoc] == Member;
      nlp = (LandParcel *)[nlpi next]) {
    LandManager *nlm = [nlp getLandManager];
    if(![l contains: nlm]) {	// We should not need to check if l contains
				// self here, because getPhysicalNeighbourList
				// will not include any of self's land parcels.
      [l addLast: nlm];
    }
  }
  [nlpi drop];
  [nlpl drop];
}

/*

getPhysicalNeighbourList:

You can ask a land manager for a list of the land parcels that are physical
neighbours of the land parcels owned. This will not include the list of land
parcels owned. The argument is a pre-allocated list pointer.

*/

-(void)getPhysicalNeighbourList: (id <List>)l {
  id <Index> lpi, nlpi;
  LandParcel *lp;

  for(lpi = [landParcelsOwned begin: scratchZone],
	lp = (LandParcel *)[lpi next];
      [lpi getLoc] == Member;
      lp = (LandParcel *)[lpi next]) {
    LandParcel *nlp;

    for(nlpi = [lp nbrBegin: scratchZone], nlp = (LandParcel *)[nlpi next];
	[nlpi getLoc] == Member;
	nlp = (LandParcel *)[nlpi next]) {
      if(![l contains: nlp] && ![landParcelsOwned contains: nlp]) {
	[l addLast: nlp];
      }
    }
    [nlpi drop];
  }
  [lpi drop];
}
  

/*

allocateLandUses

This is the method called from the schedule telling the land manager
to choose the land uses for the land parcels owned. For each land
parcel, a decision method is chosen, depending on context. If this is
the first time the land manager has the land parcel, then the
initialStrategy is used. If the yield of the land parcel is at or
above the aspirationThreshold the atOrAboveStrategy is
used. Otherwise, if a randomly chosen number between 0 and 1 is less
than or equal to pImitative, then use the
belowThresholdImitativeStrategy, else use the
belowThresholdNonImitative Strategy.

*/

-(void)allocateLandUses {
  id lpi;
  LandParcel *lp;
  LandUse *lu;
  const char *method;

  // Iterate through the land parcels owned by the land manager
  for(lpi = [landParcelsOwned begin: scratchZone], [lpi next];
      [lpi getLoc] == Member;
      [lpi next]) {
    lp = [lpi get];		// Get the land parcel from the list
    if([FearlusArguments debugNeighbourhood]) {
      [[environment getLandParcels] forEach: M(resetNImitations)];
				// Reset the counter of
				// imitation of land parcels
    }
    if([[Interval castInt: [lp getLastYield]] lt: aspirationThreshold] &&
       [lp getLastYield] != (unsigned)-1) {
      // Yield below threshold. The LandParcel
      // getLastYield method returns (unsigned)-1 if there is no last year from
      // which to get the yield.
      Interval *choice = [Interval create];

      [choice set: [Interval
		     constant:
		       [uniformDblRand getDoubleWithMin: 0.0 withMax: 1.0]]];
      if([choice le: pImitative]) {
	// Use imitative strategy
	method = class_get_class_name([belowThresholdImitativeStrategy class]);
	if([FearlusArguments debugCoreMethods]) {
	  printf("M: %s: Allocation method selected for land parcel %u at (%d,"
		 " %d): %s by land manager %u because land parcel's yield %u "
		 "is below threshold %s and choice %s made to use imitative "
		 "strategy (probability %s)\n", sel_get_name(_cmd),
		 [lp getPIN], [lp getX], [lp getY], method, pin,
		 [lp getLastYield], [aspirationThreshold printf: "%g"],
		 [choice printf: "%g"],
		 [pImitative printf: "%g"]);
	  [Interval printfclean];
	}
	lu = [belowThresholdImitativeStrategy decideLandUseForParcel: lp];
	[lp setStrategy: belowThresholdImitativeStrategy];
      }
      else {
	// Use non-imitative strategy
	method = class_get_class_name([belowThresholdNonImitativeStrategy
					class]);
	if([FearlusArguments debugCoreMethods]) {
	  printf("M: %s Allocation method selected for land parcel %u at (%d,"
		 " %d): %s by land manager %u because land parcel's yield %u "
		 "is below threshold %s and choice %s made to use non-"
		 "imitative strategy (probability %s)\n", sel_get_name(_cmd),
		 [lp getPIN], [lp getX], [lp getY], method, pin,
		 [lp getLastYield], [aspirationThreshold printf: "%g"],
		 [[choice diff: pImitative] printf: "%g"],
		 [[[Interval constant: 1.0] diff: pImitative] printf: "%g"]);
	  [Interval printfclean];
	}
	lu = [belowThresholdNonImitativeStrategy decideLandUseForParcel: lp];
	[lp setStrategy: belowThresholdNonImitativeStrategy];
      }
      [choice drop];
    }
    else {
      // Above threshold strategy (aspiration level achieved)
      method = class_get_class_name([atOrAboveThresholdStrategy class]);
      if([FearlusArguments debugCoreMethods]) {
	printf("M: %s: Allocation method selected for land parcel %u at (%d, "
	       "%d): %s by land manager %u because the "
	       "yield %u is at or above the threshold %s\n",
	       sel_get_name(_cmd), [lp getPIN], [lp getX], [lp getY], method,
	       pin, [lp getLastYield], [aspirationThreshold printf: "%g"]);
      }
      lu = [atOrAboveThresholdStrategy decideLandUseForParcel: lp];
      [lp setStrategy: atOrAboveThresholdStrategy];
    }
    if([FearlusArguments debugManagers]) {
      printf("M: land manager %u selected land use %u for land parcel %u at "
	     "(%d, %d) using method %s\n", pin, [lu getPIN], [lp getPIN],
	     [lp getX], [lp getY], method);
    }
    if([FearlusArguments debugNeighbourhood]) {
      int x, y;
      int ex = [environment getSizeX];
      int ey = [environment getSizeY];

      printf("M: %s: Land parcels consulted in selecting land use of "
	     "land parcel %u at (%d, %d) using %s:\n",
	     sel_get_name(_cmd), [lp getPIN], [lp getX], [lp getY], method);
      for(y = -1; y < ey; y++) {
	if(y == -1) printf(" |");
	else printf("%d|", y % 10);
	for(x = 0; x < ex; x++) {
	  if(y == -1) printf("%d", x % 10);
	  else {
	    int count = [(LandParcel *)[environment getObjectAtX: x Y: y]
				       getNImitations];
	    
	    if(count >= 10) printf(">");
	    else printf("%d", count % 10);
	  }
	}
	printf("\n");
	if(y == -1) {
	  printf("-+");
	  for(x = 0; x < ex; x++) {
	    printf("-");
	  }
	  printf("\n");
	}
      }
      printf("\n");
    }
    [lp setNextLandUseTypeTo: lu];
  }
  [lpi drop];
}

/*

allocateInitialLandUses

This method is used during the first cycle only to enable Land Managers to
allocate initial land uses to the land parcels they own.

*/

-(void)allocateInitialLandUses {
  id lpi;
  LandParcel *lp;
  LandUse *lu;
  const char *method;

  // Iterate through the land parcels owned by the land manager
  for(lpi = [landParcelsOwned begin: scratchZone], [lpi next];
      [lpi getLoc] == Member;
      [lpi next]) {
    lp = [lpi get];
    method = class_get_class_name([initialStrategy class]);
    if([FearlusArguments debugCoreMethods]) {
      printf("M: %s: Allocation method selected for land parcel %u at (%d, "
	     "%d): %s by land manager %u because it is year 0\n",
	     sel_get_name(_cmd), [lp getPIN], [lp getX],
	     [lp getY], method, pin);
    }
    lu = [initialStrategy decideLandUseForParcel: lp];
    [lp setStrategy: initialStrategy];
    if([FearlusArguments debugManagers]) {
      printf("M: land manager %u selected initial land use %u for land parcel "
             "%u at (%d, %d) using method %s\n", pin, [lu getPIN], [lp getPIN],
             [lp getX], [lp getY], method);
    }
    [lp setLandUseTypeTo: lu];
  }
  [lpi drop];
  /* Land managers can remember the climate and economy from the initial year.
     This is the only method they get called from the initialisation schedule,
     so here is where the data gets stored. Ideally, this would be done with
     a separate call from the schedule... */
  [climateMemory pushItemDrop:
		   [BitString create: [self getZone]
			      setToCopyOf: [environment getClimate]]];
  [economyMemory pushItemDrop:
		   [BitString create: [self getZone]
			      setToCopyOf: [environment getEconomy]]];
}

/*

harvest

Method called from the schedule telling land managers to harvest their yield.
Land managers get the yield from the parcel, and subtract the break-even
threshold. The difference (even if negative) is added to the wealth. The land
managers also make a note of the climate and economy at this point.

*/

-(void)harvest {
  id lpi;
  LandParcel *lp;

  for(lpi = [landParcelsOwned begin: scratchZone], [lpi next];
      [lpi getLoc] == Member;
      [lpi next]) {
    lp = [lpi get];
    if([FearlusArguments debugCoreMethods]) {
      printf("M: %s: Land manager %u harvesting yield (%u) from land parcel "
	     "%u at (%d, %d). Wealth %s will be incremented by %s ",
	     sel_get_name(_cmd), pin, [lp getYield], [lp getPIN], [lp getX],
	     [lp getY], [wealth printf: "%g"],
	     [[[Interval castInt: [lp getYield]]
		diff: [parameter breakEvenThreshold]] printf: "%g"]);
      [Interval printfclean];
    }
    [wealth add: [[Interval castInt: [lp getYield]]
			    diff: [parameter breakEvenThreshold]]];
    if([FearlusArguments debugCoreMethods]) {
      printf("to %s\n", [wealth printf: "%g"]);
      [Interval printfclean];
    }
  }
  [lpi drop];
  [climateMemory pushItemDrop:
		   [BitString create: [self getZone]
			      setToCopyOf: [environment getClimate]]];
  [economyMemory pushItemDrop:
		   [BitString create: [self getZone]
			      setToCopyOf: [environment getEconomy]]];
}

/*

sellLandParcels

Add land parcels for transfer to the list. Remove these land parcels from the
list of self's land parcels. The land parcels to transfer are the N
lowest-yielding land parcels owned by the land manager such that:

wealth + (N * [parameter landParcelPrice]) >= 0 // wealth

If a land manager must sell all their land parcels to do this, then they are
retired.

*/

-(void)sellLandParcels  {
  someLandParcelsLost = NO;
  allLandParcelsLost = NO;
  /* It is OK to set the numberOfLandParcelsGainedThisYear to zero when selling
     land parcels -- this will not wipe out any increments, because there can
     be no gaining of land parcels until it has been determined which are to
     be lost. */
  numberOfLandParcelsGainedThisYear = 0;
  // If the wealth is >= 0 we can leave this method without further ado
  if([wealth ge: [Interval constant: 0.0]]) {
    if([FearlusArguments debugManagers]) {
      printf("M: Land manager %u loses no land parcels this year since wealth "
	     "%s >= 0\n", pin, [wealth printf: "%g"]);
      [Interval printfclean];
    }
    return;
  }

  // Sort the land parcels owned by the land manager in ascending order of
  // yield. But first, shuffle them. This is to remove any bias in selling
  // land parcels with equally bad yield.
  [MiscFunc shuffleList: landParcelsOwned];
  
  [MiscFunc mergeSort: landParcelsOwned withSelector: M(getYield)];

  if([FearlusArguments debugInnerCoreMethods]) {
    id i;

    printf("M: %s: Land parcels of manager %u sorted in ascending order of "
	   "yield:\n", sel_get_name(_cmd), pin);
    for(i = [landParcelsOwned begin: scratchZone], [i next];
	[i getLoc] == Member;
	[i next]) {
      printf("\tLand parcel %u at (%d, %d) with yield %u\n", [[i get] getPIN],
	     [[i get] getX], [[i get] getY], [[i get] getYield]);
    }
    [i drop];
  }
  while([wealth lt: [Interval constant: 0.0]]
	&& [landParcelsOwned getCount] > 0) {
    LandParcel *lp;

    lp = [landParcelsOwned removeFirst];
    [landAllocator addLandParcelForSale: lp];
    if([FearlusArguments debugManagers]) {
      printf("M: Land manager %u must lose land parcel %u at (%d, %d) "
	     "with yield %u in order to make wealth %s non-negative\n", pin,
	     [lp getPIN], [lp getX], [lp getY], [lp getYield],
	     [wealth printf: "%g"]);
      [Interval printfclean];
    }
    [wealth add: [parameter landParcelPrice]];
    numberOfLandParcelsGainedThisYear--;
    someLandParcelsLost = YES;
  }
  if([FearlusArguments debugManagers]) {
    printf("M: Land manager %u has wealth %s after sale of land parcels", pin,
	   [wealth printf: "%g"]);
    [Interval printfclean];
  }
  if([landParcelsOwned getCount] == 0) {
    allLandParcelsLost = YES;
    if([FearlusArguments debugManagers]) {
      printf(" and has had to sell all their land parcels");
    }
  }
  if([FearlusArguments debugManagers]) {
    printf("\n");
  }
}

/*

getLandParcels

Return the list of land parcels owned by this land manager.

*/

-getLandParcels {
  return(landParcelsOwned);
}

/*

getWealth

Return the wealth

*/

-(Interval *)getWealth {
  return [[wealth clone] ret];
}

/*

getWealthForDisplayOnly

Return the wealth, as a double, for the sake of EZGraphing.

*/

-(double)getWealthForDisplayOnly {
  return [wealth getVar];
}

/*

getAgeAsInt

Return the age, cast into an int, for the sake of EZGraphing

*/

-(int)getAgeAsInt {
  return (int)age;
}

/*

incAge

Method called from the schedule to tell land managers to increment their age

*/

-(void)incAge {
  age++;
}

/*

getMemorySize

Return the amount of memory this land manager has

*/

-(unsigned)getMemorySize {
  return memorySize;
}

/*

recallClimateAgo:

Return the BitString of the climate remembered from n years ago.

*/

-(BitString *)recallClimateAgo: (unsigned)years {
  return (BitString *)[climateMemory getObject: years];
}

/*

recallEconomyAgo:

Return the BitString of the economy remembered from n years ago.

*/

-(BitString *)recallEconomyAgo: (unsigned)years {
  return (BitString *)[economyMemory getObject: years];
}


/*

recallYieldForParcel:yearsAgo:

Return the yield of the parcel the specified number of years ago, if it is
within memory limits (if it isn't, then the result is undefined, so beware.
If recallLandUseAgoForParcel:yearsAgo: returns nil, then the yield will not
be available.

This method is put in here because future LandManagers might not want to
divulge this information, or might want to be untruthful about it...

*/

-(unsigned)recallYieldForParcel: (LandParcel *)lp yearsAgo: (unsigned)yrs {
  return [lp recallYieldAgo: yrs];
}

/*

recallLandUseForParcel:yearsAgo:

Return the LandUse of the parcel the specified number of years ago. This is
in the LandManager object, so that LandManagers can communicate with each other
about this information (and apply any misremembering that might occur)...

If yearsAgo is too far in the past, this method will return nil.

*/

-(LandUse *)recallLandUseForParcel: (LandParcel *)lp yearsAgo: (unsigned)yrs {
  return [lp recallLandUseAgo: yrs];
}

/*

getNumberOfLandParcelsOwned

Return a count of the number of members in the list of land parcels owned by
this land manager

*/

-(int)getNumberOfLandParcelsOwned {
  return([landParcelsOwned getCount]);
}

/*

addLandParcel

Add the land parcel to the list of land parcels owned by this land manager. If
this is the first land parcel added, or the initialisation year, then there is
no cost.

*/

-(void)addLandParcel: (LandParcel *)lp {
  if([FearlusArguments debugManagers]) {
    printf("M: Adding land parcel %u at %d, %d to land mgr %u\n", [lp getPIN],
	   [lp getX], [lp getY], pin);
  }
  if(([environment getYear] > 0) && ([landParcelsOwned getCount] > 0)) {
    [wealth sub: [parameter landParcelPrice]];
    numberOfLandParcelsGainedThisYear++;
  }
  [landParcelsOwned addFirst: lp];
}

/*

updateNeighbourhoodWeighting

Update the neighbourhood weighting. This saves the current
neighbourhood weighting and any other details to be saved into the
history, and changes the neighbourhood weighting. The units are
multiplied by the number of land parcels gained.

N.B. The neighbourhood weighting should be decremented in proportion
to the number of land parcels gained, so that land managers pay *less*
attention to their neighbours when they gain land parcels.

*/

-(void)updateNeighbourhoodWeighting {
  if([FearlusArguments debugManagers]) {
    printf("M: Neighbourhood weight of land manager %u changed from %s ", pin, 
	   [neighbourWeight printf: "%g"]);
    [Interval printfclean];
  }
  [neighbourWeight sub:
		     [nbrWgtChangeUnits
		       prod: [Interval castInt:
					 numberOfLandParcelsGainedThisYear]]];

  if([neighbourWeight lt: [Interval constant: 0.0]]) {
    [neighbourWeight set: [Interval constant: 0.0]];
  }
  if([FearlusArguments debugManagers]) {
    printf("to %s\n", [neighbourWeight printf: "%g"]);
    [Interval printfclean];
  }
}

/*

recordHistory

This is a method called from the schedule at the end of a year to record the
history for that year.

*/

-(void)recordHistory {
  LandManagerHistory *newHd;

  newHd = [LandManagerHistory create: [parameter historyZone]];
  [newHd initialiseWithYear: [environment getYear]
	 landParcelsGained: numberOfLandParcelsGainedThisYear
	 neighbourhoodWeight: [neighbourWeight call]
	 wealth: [wealth call]
	 age: age];
  [history addFirst: newHd];
  if([parameter dumpHistory]
     && [parameter maxHistoryLength] < [history getCount]) {
    [[history removeLast] drop];
  }
}

/*

eligibleForNewLandParcels

A land manager is eligible for new land parcels if their wealth is greater than
or equal to the landParcelPrice.

*/

-(BOOL)eligibleForNewLandParcels {
  if([wealth ge: [parameter landParcelPrice]]) {
    // The check below is no longer needed since we're using intervals.
//     if(allLandParcelsLost) {
//       fprintf(stderr, "WARNING: A land manager with no land parcels is "
// 	      "eligible for a new land parcel -- this should never happen "
// 	      "in theory, though can occur when floating point arithmetic "
// 	      "is losing accuracy. Results from this run should not be "
// 	      "trusted.\n");
//       return(NO);
//     }
    return(YES);
  }
  else {
    return(NO);
  }
}

/*

eligibleForRemovalFromGame

A land manager is eligible for removal from the game if they have had to 
transfer all land parcels this year. Note that land managers should not be
dropped because they are stored in the history of land parcels. When printing
the land parcel history, all land managers must be available.

*/

-(BOOL)eligibleForRemovalFromGame {
  return(allLandParcelsLost);
}

/*

printHistoryToFile

Print the history of the land manager to the specified file.

*/

-(void)printHistoryToFile: (FILE *)fp {
  id indx;
  LandManagerHistory *hd;

  for(indx = [history begin: scratchZone], [indx next];
      [indx getLoc] == Member;
      [indx next]) {
    if(!alive) fprintf(fp, "Dead ");
    fprintf(fp, "Manager\t%u\t%u\t", pin, [subPop getPIN]);
    hd = [indx get];
    [hd printToFile: fp];
    fprintf(fp, "\t%s\t%s\t%s\t%s\t%s\t%s\t%s%s",
	    [neighbourMutateProb printf: "%g"],
	    [aspirationThreshold printf: "%g"], [pImitative printf: "%g"], 
	    class_get_class_name([initialStrategy class]),
	    class_get_class_name([atOrAboveThresholdStrategy class]),
	    class_get_class_name([belowThresholdImitativeStrategy class]),
	    class_get_class_name([belowThresholdNonImitativeStrategy class]),
	    [FearlusArguments nl]);
    [Interval printfclean];
  }
  [indx drop];
}

/*

getColour

Method to return the colour of the land manager. Naughty I suppose, because 
this is an observery sort of thing.

*/

-(int)getColour {
  return colour;
}

/*

getAspiration

Return the aspiration threshold

*/

-(Interval *)getAspiration {
  return [aspirationThreshold ret];
}

/*

getAspirationThresholdForDisplayOnly

Return the aspiration threshold

*/

-(double)getAspirationThresholdForDisplayOnly {
  return [aspirationThreshold getVar];
}

/*

getImitateProb

Return the probability of using an imitative strategy

*/

-(Interval *)getImitateProb {
  return [pImitative ret];
}

/*

getImitateProbForDisplayOnly

Return the probability of using an imitative strategy

*/

-(double)getImitateProbForDisplayOnly {
  return [pImitative getVar];
}

/*

getNeighbourWeight

Return the neighbourhood weighting parameter of the land manager.

*/

-(Interval *)getNeighbourWeight {
  return [[neighbourWeight clone] ret];
}

/*

getNeighbourNoise

Return the neighbourhood noise parameter of the land manager.

*/

-(Interval *)getNeighbourNoise {
  return [[neighbourMutateProb clone] ret];
}

/*

drop

Override the superclass method to drop all created items

*/

-(void)drop {
  [climateMemory resetLengthDrop: 0];
  [climateMemory drop];
  [economyMemory resetLengthDrop: 0];
  [economyMemory drop];
  [landParcelsOwned drop];
  [history deleteAll];
  [history drop];
  [neighbourWeight drop];
  [pImitative drop];
  [nbrWgtChangeUnits drop];
  [neighbourMutateProb drop];
  [aspirationThreshold drop];
  [wealth drop];
  [super drop];
}

@end
