/*
    FEARLUS model0-7-1ci: ModelSwarm.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1ci, an agent-based model of land use
    change.

    FEARLUS model0-7-1ci is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1ci is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the ModelSwarm class.

*/

#import "ModelSwarm.h"
#import "Environment.h"
#import "LandAllocator.h"
#import "FearlusArguments.h"
#import "Parameter.h"
#import "MiscFunc.h"
#import <objc/objc-api.h>

@implementation ModelSwarm

/*

createBegin

Create a modelswarm, and give it some initial values. Set up a probe for the
internal variables. I wonder why the probes should get set up here (for the
class). Surely the probes for the class should only be set up once -- not each
time the createBegin method of course. Naturally we'll only be building one
ModelSwarm, so such things don't matter, but it's a niggle.

*/

+createBegin: (id)z {
  ModelSwarm *m;
  id <ProbeMap> probeMap;

  m = [super createBegin: z];

  probeMap = [EmptyProbeMap createBegin: z];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "parameter"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "landAllocator"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "environment"
				    inClass: [self class]]];
  [probeLibrary setProbeMap: probeMap For: [self class]];

  return m;
}

/*

createEnd

Probably no need to do this, but other swarm implementations do it -- override
the createEnd method to return the result of the superclass createEnd method.

*/

-createEnd {
  return([super createEnd]);
}

/*

setParameters:

Set the parameters of the model

*/

-(void)setParameters: (Parameter *)p {
  parameter = p;
}

/*

getEnvironment

Return a pointer to the environment

*/

-(Environment *)getEnvironment {
  return(environment);
}

/*

getLandAllocator

Return a pointer to the land allocator

*/

-(LandAllocator *)getLandAllocator {
  return(landAllocator);
}

/*

initialSeed

Method called from the initialisation schedule to set the seed specified for
the run.

*/

-initialSeed {
  printf("Using seed: %u\n", [parameter specifiedSeed]);
  [randomGenerator setStateFromSeed: [parameter specifiedSeed]];
  return self;
}

/*

switchSeed

Method called from initialisation schedule if user has specified a separate
seed after initialisation.

*/

-switchSeed {
  printf("Using new seed: %u\n", [parameter postInitSeed]);
  [randomGenerator setStateFromSeed: [parameter postInitSeed]];
  return self;
}

/*

buildObjects

Override the Swarm class buildObjects method to build my own objects.
Build the environment and the land allocator. 

*/

-buildObjects {
  [super buildObjects];		// Build the super-class objects

  // Initialise/Create the sub-populations. These are stored in the parameter
  // object. They may have been created already by the user for configuration
  // -- the sub-populations objects will know if this has happened, and either
  // load themselves or not depending on whether the user has entered values
  // for them already.
  [parameter finaliseSubPopulations];

  // Initialise/Create the climate/economy toggle probability arrays. These
  // are stored in the parameter object. If I get round to doing a GUI for
  // entering them, they might have already been entered at this point.
  // Otherwise, they will get loaded in from the file
  [parameter finaliseClimateEconomyToggleProbs];

  if([FearlusArguments debugModel]) {
    printf("m: Model being built using the following parameters:\n");
    [parameter writeHistory: stdout];
  }
  // Build and initialise the environment -- this will cause the land parcels
  // to be created and clumped, the land uses created, the land uses assigned
  // to land parcels, and initial yields to be computed.

  environment = [Environment createBegin: self];
  [environment setParameters: parameter];
  environment = [environment createEnd];
  // Build and initialise the land allocator -- this will cause the land
  // managers to be created and assigned to a single land parcel each.
  if([FearlusArguments debugModel]) {
    printf("m: Creating land allocator\n");
  }
  landAllocator = [LandAllocator createBegin: self];
  [landAllocator setParameters: parameter];
  landAllocator = [landAllocator createEnd];
  return(self);
}

/*

buildActions

Override the Swarm class buildActions method to build the schedule for this
model. Things happen in the following order:

1.	The land managers determine the land use for their land parcel(s)
2.	The environment determines the climate for the year
3.	The environment determines the economy for the year
4.	The environment tells each land parcel to compute its yield
5.	The land allocator transfers land parcels between land managers
6.	The land allocator removes from the game any land managers with no land
7.	The land managers update their strategies and strategy histories
8.	The environment performs tasks to update the histories of the land
        parcels and land uses, and initiates the next year.



*/

-buildActions {
  unsigned i;

  [super buildActions];		// You have to let the super class build its
				// actions
  // Build the action group for the actions to take place cycle by cycle
  modelActions = [ActionGroup create: self];
  // The order in which the actions are created is the order in which they will
  // be executed by the scheduler.
  [modelActions createActionTo: environment message: M(newYear)];
				// Environment initiate a new year.
  for(i = 0; i < [parameter nSubPopulations]; i++) {
    [modelActions createActionTo: [parameter subPopulation: i]
		  message: M(resetRemovals)];
  }
  [modelActions createActionForEach: [environment getLandParcels]
		message: M(updateLandManager)];
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(allocateLandUses)];
				// Land managers allocate land uses
  [modelActions createActionForEach: [environment getLandParcels]
		message: M(updateLandUse)];
                                // Land parcels adopt their new land use
  [modelActions createActionTo: environment message: M(determineClimate)];
				// Environment determine the climate
  [modelActions createActionTo: environment message: M(determineEconomy)];
				// Environment determine the economy
  [modelActions createActionTo: environment message: M(calculateYield)];
				// Environment calculate the yield
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(harvest)];
                                // Land managers harvest their yield
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(sellLandParcels)];
  [modelActions createActionTo: landAllocator message: M(transferLandParcels)];
				// Land allocator transfer land parcels
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(incAge)];
                                // Land Managers increment their age
  [modelActions createActionTo: landAllocator message: M(killLosers)];
				// Land allocator kill off losing land mgrs
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(updateNeighbourhoodWeighting)];
				// Land managers update neighbourhood weighting
  [modelActions createActionTo: environment message: M(recordHistory)];
  [modelActions createActionTo: landAllocator message: M(recordHistory)];

  modelSchedule = [Schedule createBegin: self];
  [modelSchedule setRepeatInterval: 1];
 
  modelSchedule = [modelSchedule createEnd];
  [modelSchedule at: 0 createAction: modelActions];
				// Put the action group into the schedule
  // Build the action group for the actions to take place in the initial cycle
  // only
/*

Initialise the objects created with buildObjects. The environment
will build the land parcels and land uses. The land allocator will build the
land managers. At the end of the initialisation schedule, we want the following
to have happened:

- All land parcels created
- All land parcels clumped
- All land uses created
- Initial land uses assigned to land parcels
- Initial yields computed for each land parcel
- All land managers created
- All land parcels assigned to land managers

*/
  initialActions = [ActionGroup create: self];
  // Set the seed for the run. THIS MUST BE THE FIRST ACTION.
  [initialActions createActionTo: self message: M(initialSeed)];

  [initialActions createActionTo: environment message: M(createLandParcels)];
  [initialActions createActionTo: environment message: M(connectLandParcels)];
  if([parameter useLandParcelFile] 
     && [MiscFunc fileExists: [parameter landParcelFile]]) {
    [initialActions createActionTo: environment
		    message: M(loadLandParcelsFromFileNamed:)
		    : (id)[parameter landParcelFile]];
    // The cast to an id is OK. [parameter landParcelFile] returns a char *,
    // and id is really a pointer, so they are both using the same amount of
    // memory, and it stops the compiler from whinging.
  }
  else {
    [initialActions createActionTo: environment message: M(clumpLandParcels)];
    if([parameter useLandParcelFile]) {
      [initialActions createActionTo: environment
		      message: M(saveLandParcelsToFileNamed:)
		      : (id)[parameter landParcelFile]];
    }
  }
  [initialActions createActionForEach: [environment getLandParcels]
		  message: M(determineNWNeighbourAttributeMatch)];
  [initialActions createActionTo: environment message: M(createLandUses)];
  if([parameter useLandUseFile]) {
    if([MiscFunc fileExists: [parameter landUseFile]]) {
      [initialActions createActionTo: environment
		      message: M(loadLandUsesFromFileNamed:)
		      : (id)[parameter landUseFile]];
    }
    else {
      [initialActions createActionTo: environment
		      message: M(saveLandUsesToFileNamed:)
		      : (id)[parameter landUseFile]];
    }
  }
  [initialActions createActionTo: environment message: M(determineClimate)];
  [initialActions createActionTo: environment message: M(determineEconomy)];
  /* NB Climate and economy are determined before land use in the initial cycle
     only -- is this an acceptable anomaly? */
  [initialActions createActionTo: landAllocator
		  message: M(initialiseWithEnvironment:) : environment];
  [initialActions createActionForEach: [landAllocator getLandManagers]
		  message: M(allocateInitialLandUses)];
  [initialActions createActionTo: environment message: M(calculateYield)];
  [initialActions createActionTo: environment message: M(recordHistory)];
  [initialActions createActionTo: landAllocator message: M(recordHistory)];

  // If a separate seed is required after initialisation, then add an action
  // at the end of the initialisation phase to set that seed
  // THIS MUST BE THE LAST ACTION
  if([FearlusArguments postInitSeedHasBeenSpecified]) {
    [initialActions createActionTo: self message: M(switchSeed)];
  }
  //  [initialActions createActionTo: modelSchedule 
  //	  message: M(activateIn:) : self];
  // Build the schedule
  initialSchedule = [Schedule createBegin: self];
  initialSchedule = [initialSchedule createEnd];
  [initialSchedule at: 0 createAction: initialActions];
  [initialSchedule at: 1 createActionTo: modelSchedule
		   message: M(activateIn:) : self];
  return(self);
}

/*

activateIn

Activate this model swarm in the required context.



*/

-activateIn: (id)swarmContext {
  [super activateIn: swarmContext];
  [initialSchedule activateIn: self];
  //  [modelSchedule activateIn: self];
  return([self getSwarmActivity]);
}

/*

writeHistory

Write the history of the model to a file passed as an argument to this method

*/

-(void)writeHistory: (FILE *)fp {
  // If you do ever put anything written to fp by the modelswarm directly in
  // here, then remember to use [FearlusArguments nl] to get the new line string.
  [parameter writeHistory: fp];
  [environment writeHistory: fp];
  [landAllocator writeHistory: fp];
}

/*

drop

Drop any object created by this object

*/

-(void)drop {
  [environment drop];
  [landAllocator drop];
  [super drop];
}

@end
