/*
    FEARLUS model0-7-1ci: Parameter.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1ci, an agent-based model of land use
    change.

    FEARLUS model0-7-1ci is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1ci is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/


#import "Parameter.h"
#import "SubPopulation.h"
#import "MiscFunc.h"
#import "NumberArray.h"
#import "FearlusArguments.h"
#import "Environment.h"
#import "StrategyColourKey.h"
#import "Interval.h"
#import <string.h>
#import <sys/types.h>
#import <unistd.h>
#import <pwd.h>
#import <sys/utsname.h>
#import <time.h>

#define EOFCHECK(expr, filen, file, quit) if(!(expr)) { \
  fprintf(stderr, "%s -- Unexpected EOF in file %s\n", sel_get_name(_cmd), \
          (filen)); \
  [(file) drop]; \
  if((quit)) exit(1); else return; \
}

@implementation Parameter

+createBegin: (id)z {
  Parameter *p;
  id <ProbeMap> pmap;

  p = [super createBegin: z];
  p->setData = NO;
  p->environmentType = "Toroidal-VonNeumann";
  p->neighbourhoodRadius = 1;
  p->neighbourhoodIsMoore = -1;
  p->environmentIsBounded = -1;
  p->clumping = "None";
  p->neighbourNoiseMax = 0.0;
  p->neighbourNoiseMin = 0.0;
  p->strategyChangeUnit = 0.0;
  p->useLandParcelFile = NO;
  p->useLandUseFile = NO;
  p->useClimateFile = NO;
  p->useEconomyFile = NO;
  p->subPopArrayCreated = NO;
  p->climEconCreated = NO;
  p->panelCreated = NO;
  p->subPopFile = "fearlus.sp";
  p->climateToggleProbFile = "fearlus.ct";
  p->economyToggleProbFile = "fearlus.et";
  p->dumpHistory = YES;
  p->maxHistoryLength = 500;
  p->infiniteTime = YES;
  p->maxYear = 0;
  p->nInitXParcels = 1;
  p->nInitYParcels = 1;
  p->historyZone = [Zone create: globalZone];

  pmap = [DefaultProbeMap createBegin: z];
  [pmap setProbedClass: [self class]];
  pmap = [pmap createEnd];
  [pmap dropProbeForVariable: "clumperobj"];
  [pmap dropProbeForVariable: "nLandParcelClump"];
  [pmap dropProbeForVariable: "neighbourhoodIsMoore"];
  [pmap dropProbeForVariable: "environmentIsBounded"];
  [pmap dropProbeForVariable: "setData"];
  [pmap dropProbeForVariable: "panelCreated"];
  [pmap dropProbeForVariable: "climEconCreated"];
  [pmap dropProbeForVariable: "subPopArrayCreated"];
  [pmap dropProbeForVariable: "dumpHistory"];
  [pmap dropProbeForVariable: "maxHistoryLength"];
  [pmap dropProbeForVariable: "maxMemorySize"];
  [pmap dropProbeForVariable: "strategyKey"];
  [pmap dropProbeForVariable: "historyZone"];
  [pmap addProbe: [probeLibrary getProbeForMessage:
				  "clickMeToGetSubPopulationArray"
				inClass: [self class]]];

  [probeLibrary setProbeMap: pmap For: [self class]];

  return p;
}

-createEnd {
  return [super createEnd];
}

/*

finalise

Set up the environment class and clumper. These will not be changeable
henceforth. They should also not be used until this method has been
run.

*/

-finalise {
  Class clumperClass;
  char *clumpStr, *clumpClassStr, *clumpParamStr, *clumpParamArgStr;

  setData = YES;

  /* Deal with deprecated environment invokations */

  if(neighbourhoodIsMoore != -1 || environmentIsBounded != -1) {
    fprintf(stderr, "WARNING: Deprecated environment invokation in model file "
	    "-- use environmentType instead of neighbourhoodIsMoore and "
	    "environmentIsBounded\n");
    if(neighbourhoodIsMoore == 0 && environmentIsBounded == 0) {
      environmentType = "Toroidal-VonNeumann";
    }
    else if(neighbourhoodIsMoore == 0 && environmentIsBounded == 1) {
      environmentType = "Planar-VonNeumann";
    }
    else if(neighbourhoodIsMoore == 1 && environmentIsBounded == 0) {
      environmentType = "Toroidal-Moore";
    }
    else if(neighbourhoodIsMoore == 1 && environmentIsBounded == 1) {
      environmentType = "Planar-Moore";
    }
    else {
      fprintf(stderr, "ERROR: Cannot determine environmentType from "
	      "neighbourhoodIsMoore = %d and environmentIsBounded = %d\n",
	      neighbourhoodIsMoore, environmentIsBounded);
      exit(1);
    }
    fprintf(stderr, "WARNING: Setting environmentType to %s\n", 
	    environmentType);
  }

  /* Set up the clumper */

  clumpStr = strdup(clumping);
  clumpClassStr = clumpStr;
  if(strcmp(clumpClassStr, "None") == 0) {
    clumperobj = nil;
    clumperClass = Nil;
  }
  else {
    /* The user has specified a clumper class. This is written as a string
       with a colon separating the clumper class on the LHS with a semi-colon
       separated list of clumperclass specific parameters on the RHS. Each
       parameter is written PARAM=VALUE. For example:

       CAVNT3Clumper:cycles=4		// One parameter
       None			    	// No clumping (dealt with above)
       Same10PropClumper		// A clumper not requiring arguments
       ManyParamClumper:p1=4.5;p2=yes	// A clumper with many parameters

    */
    char *p, *q;

    p = strchr(clumpStr, (int)':');	// Look for the colon
    if(p != NULL) {
      *p = '\0';			// Replace the : with a terminator
    }

    /* Now we know the clumper class string, so create the clumper */
    clumperClass = objc_get_class(clumpClassStr);
    clumperobj = [clumperClass create: [self getZone]];
    [clumperobj setParameters: self];
    if(![clumperobj conformsTo: @protocol(Clumper)]) {
      fprintf(stderr, "Clumper class %s does not conform to Clumper "
	      "protocol\n", clumpClassStr);
      exit(1);
    }
    while(p != NULL) {			// If there was a : then do the params
      p++;
      clumpParamStr = p;
      q = strchr(p, (int)'=');		// Look for the =
      if(q == NULL) {
	fprintf(stderr, "Error in clump parameter string: Found a clump "
		"parameter %s, but can't find an =\n", p);
	exit(1);
      }
      *q = '\0';			// Replace the = with a terminator
      q++;
      clumpParamArgStr = q;
      p = strchr(q, (int)';');		// Look for a ; if any
      if(p != NULL) {
	*p = '\0';			// Replace the ; with a terminator
      }
      /* Pass the parameter into the clumper object */
      if(![clumperobj setParameter: clumpParamStr toValue: clumpParamArgStr]) {
	fprintf(stderr, "Error in clump parameter string: Invalid parameter "
		"or invalid parameter value for Clumper class %s: %s=%s\n",
		clumpClassStr, clumpParamStr, clumpParamArgStr);
	exit(1);
      }
    }
  }
  free(clumpStr);
  /* Done setting up the clumper */
  return self;
}

-setEnvironmentType: (char *)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  environmentType = strdup(value);
  return self;
}

-setNeighbourhoodRadius: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  neighbourhoodRadius = value;
  return self;
}

-setMooreNeighbourhood: (BOOL)value {
  fprintf(stderr, "%s -- Warning: This method is deprecated\n",
	  sel_get_name(_cmd));
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  neighbourhoodIsMoore = value;
  return self;
}

-setEnvironmentBounded: (BOOL)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  environmentIsBounded = value;
  return self;
}

-setClimateBitStringSize: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  climateBSSize = value;
  return self;
}

-setEconomyBitStringSize: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  economyBSSize = value; 
  return self;
}

-setParcelBitStringSize: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  landParcelBSSize = value;
  return self;
}

-setNUses: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  nLandUse = value;
  return self;
}

-setPUseDontCareBit: (double)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  pLandUseDontCare = value;
  return self;
}

-setNParcelClumpCycles: (unsigned)value {
  fprintf(stderr, "%s -- Warning: This method is deprecated\n",
	  sel_get_name(_cmd));
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  nLandParcelClump = value;
  return self;
}

-setClumper: (id <Clumper>)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  clumperobj = value;
  return self;
}

-setEnvironmentXSize: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  envXSize = value;
  return self;
}

-setEnvironmentYSize: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  envYSize = value;
  return self;
}

-setStrategyChangeUnit: (double)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  strategyChangeUnit = value;
  if(value != 0.0) {
    fprintf(stderr, "%s -- WARNING: strategyChangeUnit (%g) out of range for "
	    "imitation experiments (should be 0.0)\n", sel_get_name(_cmd),
	    value);
  }
  return self;
}

-setNeighbourNoiseMax: (double)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  neighbourNoiseMax = value;
  if(value != 0.0) {
    fprintf(stderr, "%s -- WARNING: neighbourNoiseMax (%g) out of range for "
	    "imitation experiments (should be 0.0)\n", sel_get_name(_cmd),
	    value);
  }
  return self;
}

-setNeighbourNoiseMin: (double)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  neighbourNoiseMin = value;
  if(value != 0.0) {
    fprintf(stderr, "%s -- WARNING: neighbourNoiseMin (%g) out of range for "
	    "imitation experiments (should be 0.0)\n", sel_get_name(_cmd),
	    value);
  }
  return self;
}

-setBreakEvenThreshold: (double)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  breakEvenThreshold = value;
  return self;
}

-setLandParcelPrice: (double)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  landParcelPrice = value;
  return self;
}

-setNInitXParcels: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  nInitXParcels = value;
  return self;
}

-setNInitYParcels: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  nInitYParcels = value;
  return self;
}

/*

setDumpHistory:

This method permits the parameter to be set after createEnd -- because users
will set their values in the ObserverSwarm probe after the parameters have
been loaded. The ObserverSwarm will then needs to set the values in the
parameter object.

*/

-setDumpHistory: (BOOL)value {
  dumpHistory = value;
  return self;
}

/*

setMaxHistoryLength:

For much the same reasons as setDumpHistory: this method does not check the
setData flag.

*/

-setMaxHistoryLength: (unsigned)value {
  maxHistoryLength = value;
  return self;
}

-setLandParcelFile: (char *)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  landParcelFile = value;
  useLandParcelFile = YES;
  return self;
}

-setLandUseFile: (char *)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  landUseFile = value;
  useLandUseFile = YES;
  return self;
}

-setClimateFile: (char *)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  climateFile = value;
  useClimateFile = YES;
  return self;
}

-setEconomyFile: (char *)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  economyFile = value;
  useEconomyFile = YES;
  return self;
}

-setClimateToggleProbFile: (char *)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  climateToggleProbFile = value;
  return self;
}

-setEconomyToggleProbFile: (char *)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  economyToggleProbFile = value;
  return self;
}

-setNSubPopulations: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  nSubPops = value;
  return self;
}

-setSubPopulationFile: (char *)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  subPopFile = value;
  return self;
}

-setMaximumYear: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  maxYear = value;
  return self;
}

-setInfiniteTime: (BOOL)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  infiniteTime = value;
  return self;
}

-setSpecifiedSeed: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  seed = value;
  return self;
}

-setPostInitSeed: (unsigned)value {
  if(setData) {
    fprintf(stderr, "%s -- Parameter already set. Cannot change it\n",
	    sel_get_name(_cmd));
    return self;
  }
  postInitSeed = value;
  return self;
}

-setStrategyColourKey: (StrategyColourKey *)key {
  strategyKey = key;
  return self;
}

-(char *)environmentType {
  return environmentType;
}

-(unsigned)neighbourhoodRadius {
  return neighbourhoodRadius;
}

-(BOOL)mooreNeighbourhood {
  fprintf(stderr, "%s -- Warning: This method is deprecated\n",
	  sel_get_name(_cmd));
  return neighbourhoodIsMoore;
}

-(BOOL)environmentBounded {
  return environmentIsBounded;
}

-(unsigned)climateBitStringSize { 
  return climateBSSize;
}

-(NumberArray *)climateToggleProbs {
  return climateToggleProbs;
}

-(unsigned)economyBitStringSize {
  return economyBSSize;
}

-(NumberArray *)economyToggleProbs {
  return economyToggleProbs;
}

-(unsigned)parcelBitStringSize {
  return landParcelBSSize;
}

-(unsigned)useBitStringSize {
  return climateBSSize + economyBSSize + landParcelBSSize;
}

-(unsigned)nUses { 
  return nLandUse;
}

-(Interval *)pUseDontCareBit { 
  return [[Interval constant: pLandUseDontCare] call];
}

-(id <Clumper>)clumper {
  return clumperobj;
}

-(unsigned)nParcelClumpCycles { 
  fprintf(stderr, "%s -- Warning: This method is deprecated\n",
	  sel_get_name(_cmd));
  return nLandParcelClump;
}

-(unsigned)environmentXSize { 
  return envXSize;
}

-(unsigned)environmentYSize { 
  return envYSize;
}

-(Interval *)strategyChangeUnit { 
  return [[Interval constant: strategyChangeUnit] ret];
}

-(double)neighbourNoiseMax { 
  return neighbourNoiseMax;
}

-(double)neighbourNoiseMin { 
  return neighbourNoiseMin;
}
-(unsigned)maximumMemorySize {
  return maxMemorySize;
}

-(Interval *)breakEvenThreshold { 
  return [[Interval constant: breakEvenThreshold] ret];
}

-(Interval *)landParcelPrice { 
  return [[Interval constant: landParcelPrice] ret];
}

-(unsigned)nInitXParcels { 
  return nInitXParcels;
}

-(unsigned)nInitYParcels { 
  return nInitYParcels;
}

-(BOOL)dumpHistory {
  return dumpHistory;
}

-(unsigned)maxHistoryLength {
  return maxHistoryLength;
}

-(id <Zone>)historyZone {
  return historyZone;
}

-(char *)landParcelFile {
  return landParcelFile;
}

-(BOOL)useLandParcelFile {
  return useLandParcelFile;
}

-(char *)landUseFile {
  return landUseFile;
}

-(BOOL)useLandUseFile {
  return useLandUseFile;
}

-(char *)climateFile {
  return climateFile;
}

-(BOOL)useClimateFile {
  return useClimateFile;
}

-(char *)economyFile {
  return economyFile;
}

-(BOOL)useEconomyFile {
  return useEconomyFile;
}

-(char *)climateToggleProbFile {
  return climateToggleProbFile;
}

-(char *)economyToggleProbFile {
  return economyToggleProbFile;
}

-(unsigned)nSubPopulations {
  return nSubPops;
}

-(SubPopulation *)subPopulation: (unsigned)i {
  return [subPops atOffset: (int)i];
}

-(BOOL)infiniteTime {
  return infiniteTime;
}

-(unsigned)maximumYear {
  return maxYear;
}

-(unsigned)specifiedSeed {
  return seed;
}

-(unsigned)postInitSeed {
  return postInitSeed;
}

-(StrategyColourKey *)strategyColourKey {
  return strategyKey;
}

/* 

loadFromFileNamed:

Load in the parameters from the named file. There's a bit of tidying
up to do after loading the stuff in, so we can't use ObjectLoader
directly from Batch/ObserverSwarm any more.

*/

-loadFromFileNamed: (const char *)filename {
  [ObjectLoader load: self fromFileNamed: filename];
  return self;
}

/*

writeHistory:

Write the parameters to the specified history file pointer. Call each of the
sub-populations to get them to write their history too.

*/

-(void)writeHistory: (FILE *)fp {
  int i;
  const char *nl;
  struct passwd *rpwent, *epwent, pwnotfound;
  struct utsname minfo;
  const time_t t = time(NULL);
  struct tm *lt = localtime(&t);

  nl = [FearlusArguments nl];
  
  pwnotfound.pw_name = "unknown!";
  rpwent = getpwuid(getuid());
  if(rpwent == NULL) rpwent = &pwnotfound;
  epwent = getpwuid(geteuid());
  if(epwent == NULL) epwent = &pwnotfound;
  if(uname(&minfo) < 0) {
    strcpy(minfo.sysname, "unknown!");
    strcpy(minfo.nodename, "unknown!");
    strcpy(minfo.release, "");
    strcpy(minfo.version, "");
    strcpy(minfo.machine, "unknown!");
  }
  fprintf(fp, "Version:\t%s%s", [arguments getAppName], nl);
  fprintf(fp, "Path:\t%s%s", [arguments getExecutablePath], nl);
  fprintf(fp, "Swarm:\t%s%s", [arguments getSwarmHome], nl);
  fprintf(fp, "User:\tReal:\t%s\t", rpwent->pw_name);
  fprintf(fp, "Effective:\t%s%s", epwent->pw_name, nl);
  fprintf(fp, "Machine:\tName:\t%s\tOS:\t%s %s %s\tArchitecture:\t%s%s",
	  minfo.nodename, minfo.sysname, minfo.release, minfo.version,
	  minfo.machine, nl);
  fprintf(fp, "Date:\t%04d/%02d/%02d %02d:%02d:%02d%s", lt->tm_year + 1900,
	  lt->tm_mon + 1, lt->tm_mday, lt->tm_hour, lt->tm_min,
	  lt->tm_sec, nl);
  if([FearlusArguments seedHasBeenSpecified]) {
    fprintf(fp, "Seed for simulation:\t%u%s", seed, nl);
  }
  else {
    fprintf(fp, "Seed for simulation:\t%u\t<Swarm>%s", seed, nl);
  }
  if([FearlusArguments postInitSeedHasBeenSpecified]) {
    fprintf(fp, "Separate seed for use after initialisation:\t%u%s",
            postInitSeed, nl);
  }
  else {
    fprintf(fp, "Separate seed for use after initialisation:\t<Not Used>%s",
	    nl);
  }
  fprintf(fp, "Termination year:\t%u%sUnbounded time:\t%s%s",
	  maxYear, nl, (infiniteTime ? "YES" : "NO"), nl);
  fprintf(fp, "Environment type:\t%s%s"
	  "Climate BitString size:\t%u%sEconomy BitString size:\t%u%s"
	  "Land Parcel BitString size:\t%u%sLand Use BitString size:\t%u%s"
	  "Number of Land Uses:\t%u%sProbability of Don't Care bit in Land Use"
	  " BitString:\t%g%sNumber of clumping cycles in Land Parcel BitString"
	  " clumping CA:\t%u%sNumber of horizontal cells in the Environment:"
	  "\t%u%sNumber of vertical cells in the Environment:\t%u%s"
	  "Unit of change of strategy:\t%g%s"
	  "Maximum noise when Land Manager gets neighbouring Land Parcel "
	  "BitString:\t%g%sMinimum noise when Land Manager gets neighbouring "
	  "Land Parcel BitString:\t%g%sBreak even yield threshold:\t%g%s"
	  "Fixed Land Parcel sale price:\t%g%s"
	  "Number of X initial Parcels for Managers:\t%u%s"
	  "Number of Y initial Parcels for Managers:\t%u%s",
	  environmentType, nl,
	  climateBSSize, nl,
	  economyBSSize, nl, landParcelBSSize, nl,
	  climateBSSize + economyBSSize + landParcelBSSize, nl,
	  nLandUse, nl, pLandUseDontCare, nl, nLandParcelClump, nl, envXSize,
	  nl, envYSize, nl, strategyChangeUnit, nl, neighbourNoiseMax, nl,
	  neighbourNoiseMin, nl, breakEvenThreshold, nl, landParcelPrice, nl,
	  nInitXParcels, nl, nInitYParcels, nl);
  fprintf(fp, "Dump history:\t%s%sMaximum history length:\t%u%s",
	  (dumpHistory ? "YES" : "NO"), nl, maxHistoryLength, nl);
  fprintf(fp, "Land Parcel file:\t%s%sLand Use file:\t%s%sClimate file:\t%s%s"
	  "Economy file:\t%s%s",
	  (useLandParcelFile ? landParcelFile : "Not used"), nl,
	  (useLandUseFile ? landUseFile : "Not used"), nl,
	  (useClimateFile ? climateFile : "Not used"), nl,
	  (useEconomyFile ? economyFile : "Not used"), nl);
  fprintf(fp, "Climate toggle probabilities:\t");
  for(i = 0; i < (int)climateBSSize; i++) {
    fprintf(fp, "%s%s", [[climateToggleProbs getIDoubleAtOffset: i]
			  printf: "%g"],
	    ((i == climateBSSize - 1) ? nl : "\t"));
  }
  if(climateBSSize == 0) {
    fprintf(fp, "None (Zero length bit-string)%s", nl);
  }
  fprintf(fp, "Economy toggle probabilities:\t");
  for(i = 0; i < (int)economyBSSize; i++) {
    fprintf(fp, "%s%s", [[economyToggleProbs getIDoubleAtOffset: i]
			  printf: "%g"],
	    ((i == economyBSSize - 1) ? nl : "\t"));
  }
  if(economyBSSize == 0) {
    fprintf(fp, "None (Zero length bit-string)%s", nl);
  }
  [Interval printfclean];
  for(i = 0; i < (int)nSubPops; i++) {
    [[subPops atOffset: i] writeHistory: fp];
  }
}

/*

clickMeToGetSubPopulationArray

This method is provided so that people can click on the Parameter probe map
and look at/edit the parameters for the sub-populations. It creates a button
panel object, with one button for each sub-population, which, when clicked,
causes the sub-population to display its probe map (hopefully).

*/

-clickMeToGetSubPopulationArray {
  int i;

  if(panelCreated) {
    fprintf(stderr, "Cannot create sub-population button panel because it "
	    "has already been created.\n");
    /* It would be nice if this had some nicer front end. E.g. I think
     there is a "destroy notification" in the Swarm libraries, so, if
     the panel had been destroyed, then it would let you recreate
     it. Otherwise, perhaps it could deiconify the button panel and
     bring it to the foreground. */
    return self;
  }
  panelCreated = YES;
  if(!subPopArrayCreated) {
    [self createSubPopulations];
    [self loadSubPopulations: NO];
  }
  panel = [ButtonPanel create: [self getZone]];
  for(i = 0; i < (int)nSubPops; i++) {
    [panel addButtonName: [[subPops atOffset: i] getLabel]
	   target: [subPops atOffset:i]
	   method: M(displayProbe)];
  }
  [panel setWindowTitle: "Sub-Pops"];
  return self;
}

/*

createSubPopulations

Create the sub-populations array if it hasn't been created already, with
nSubPops elements. Basically, at the end of this method, nSubPops and subPops
will correspond and subPops will be accessible.

*/

-(void)createSubPopulations {
  if(!subPopArrayCreated) {
    int i;

    subPops = [Array create: [self getZone] setCount: (int)nSubPops];
    for(i = 0; i < nSubPops; i++) {
      SubPopulation *sp;

      sp = [SubPopulation create: [self getZone]];
      [subPops atOffset: i put: sp];
    }
    subPopArrayCreated = YES;
  }
  else {
    nSubPops = (unsigned)[subPops getCount];
  }
}

/*

loadSubPopulations

If load the sub-populations in from the sub-population file. There is a flag
passed in as argument:

If YES, then (a) Quit if the sub-population file cannot be found; (b) take
the size of the sub-population array from the file; and (c) create the 
sub-population array

If NO, then (a) Do not quit under any circumstances; (b) do not load the data
in from the file if the required size of sub-population array is different from
the currently set value; and (c) create the sub-population array if it hasn't
been created already.

Basically NO is from GUI mode, YES is from creation from the modelSwarm.

File format:

NumberOfSubPopulations: nn
SubPopulationFile 1: ffffffff.01.sp
SubPopulationFile 2: ffffffff.02.sp
...
SubPopulationFile nn: ffffffff.nn.sp

This file contains a list of files from which to load each
sub-population's data.

*/

-(void)loadSubPopulations: (BOOL)quitFlag {
  int i;
  unsigned dummyNSubPops;
  id file;
  char buf[1024];

  if(subPopArrayCreated) {
    nSubPops = (unsigned)[subPops getCount];
    // Fix up the nSubPops parameter in case anyone has changed it
    // Aargh they will always be able to change it during the run -- keep quiet
    // about this!
  }
  file = [InFile create: [self getZone] setName: subPopFile];
  if(file == nil) {
    if(!quitFlag) {
      return;
    }
    fprintf(stderr, "%s -- could not open file ", sel_get_name(_cmd));
    perror(subPopFile);
    exit(1);
  }
  EOFCHECK([file getWord: buf], subPopFile, file, quitFlag)
    if(strcmp(buf, "NumberOfSubPopulations:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), subPopFile,
	      "NumberOfSubPopulations:", buf);
      [file drop];
      if(quitFlag) exit(1); else return;
    }
  EOFCHECK([file getUnsigned: &dummyNSubPops], subPopFile, file, quitFlag)
  if(dummyNSubPops != nSubPops) {
    if(quitFlag) {
      if(subPopArrayCreated) {
	fprintf(stderr, "%s -- cannot load sub-population data in from file %s"
		" -- sub-population array already created, and number of sub-"
		"populations in the file (%u) is not equal to the number of "
		"sub-populations in the array (%u)\n", sel_get_name(_cmd),
		subPopFile, dummyNSubPops, nSubPops);
	[file drop];
	exit(1);
      }
    }
    else {
      [file drop];
      return;
    }
  }
  if(quitFlag) {
    nSubPops = dummyNSubPops;
    [self createSubPopulations];
  }
  // By this point, there must be a created array of sub-populations, with
  // nSubPops set to the number of sub-populations in the array and in the file
  for(i = 0; i < (int)nSubPops; i++) {
    EOFCHECK([file getWord: buf], subPopFile, file, quitFlag)
      if(strcmp(buf, "SubPopulationFile") != 0) {
	fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
		"%s\n", sel_get_name(_cmd), subPopFile, "SubPopulationFile",
		buf);
	[file drop];
	if(quitFlag) exit(1); else return;
      }
    EOFCHECK([file getWord: buf], subPopFile, file, quitFlag)
      if(atoi(buf) != i + 1 || buf[strlen(buf) - 1] != ':') {
	fprintf(stderr, "%s -- invalid format in file %s. Expecting %s %d: "
		"found %s %s\n", sel_get_name(_cmd), subPopFile,
		"SubPopulationFile", i, "SubPopulationFile", buf);
	[file drop];
	if(quitFlag) exit(1); else return;
      }
    EOFCHECK([file getWord: buf], subPopFile, file, quitFlag)
    [[subPops atOffset: i] loadFromFileNamed: buf];
                                  	// Load the sub-population data for
                                        // sub-population i
  }
  [file drop];
}

/*

finaliseSubPopulations

This method is called from the modelSwarm when it's time to finalise the 
SubPopulations. At this point, there are two possibilities:

1. The ClickMeToGetSubPopulationArray button was not pressed. If this is the
   case, then the sub-population array has not been created, and thus contains
   no data. The sub-population file should have been set, since this is where
   the data will be loaded from after creation.

2. The ClickMeToGetSubPopulationArray button was pressed. The sub-population
   array has been created, loaded from a file and then possibly editted. There
   is thus nothing to do.

*/

-(void)finaliseSubPopulations {
  int i;

  if(subPopArrayCreated) {
    return;
  }
  [self createSubPopulations];
  [self loadSubPopulations: YES];

  /* Compute the maximum memory size */

  maxMemorySize = 0;
  for(i = 0; i < nSubPops; i++) {
    unsigned subPopMemSz 
      = [(SubPopulation *)[subPops atOffset: i] getMaxMemorySize];

    if(subPopMemSz > maxMemorySize) maxMemorySize = subPopMemSz;
  }
}

/*

createClimateEconomyToggleProbs

This creates and initialises the climate and economy toggle probability arrays.
At the end of this method, the lengths of the toggle probability arrays and
the bit strings they correspond to will be the same, and all bits will be set
to 0.5 (the default).

*/

-(void)createClimateEconomyToggleProbs {
  int i;

  if(!climEconCreated) {
    climateToggleProbs = [NumberArray create: [self getZone]
				      setCount: climateBSSize];
    economyToggleProbs = [NumberArray create: [self getZone]
				      setCount: economyBSSize];
    climEconCreated = YES;
  }
  else {
    [climateToggleProbs setCount: climateBSSize];
    [economyToggleProbs setCount: economyBSSize];
  }
  for(i = 0; i < climateBSSize; i++) {
    [climateToggleProbs atOffset: i setIDouble:
			  [[Interval constant: 0.5] call]];
  }
  for(i = 0; i < economyBSSize; i++) {
    [economyToggleProbs atOffset: i setIDouble:
			  [[Interval constant: 0.5] call]];
  }
}

/*

loadClimateEconomyToggleProbs:

Load in the toggle probabilities. If the quitFlag is set to YES, then quit
if the file is not found, otherwise just return. Also, if the quitFlag is set
to YES, then quit if the number of items in either of the toggleFiles is not
equal to the corresponding bit-string size.

*/

-(void)loadClimateEconomyToggleProbs: (BOOL)quitFlag {
  if(![MiscFunc fileExists: climateToggleProbFile] && !quitFlag) {
    return;
  }
  [climateToggleProbs loadFromFileNamed: climateToggleProbFile];
  if((unsigned)[climateToggleProbs getCount] != climateBSSize) {
    if(quitFlag) {
      fprintf(stderr, "%s -- Climate toggle probability file %s number of "
	      "elements (%d) is not equal to climate bit string size (%u)\n",
	      sel_get_name(_cmd), climateToggleProbFile,
	      [climateToggleProbs getCount], climateBSSize);
      exit(1);
    }
    else {
      climateBSSize = (unsigned)[climateToggleProbs getCount];
    }
  }
  if(![MiscFunc fileExists: economyToggleProbFile] && !quitFlag) {
    return;
  }
  [economyToggleProbs loadFromFileNamed: economyToggleProbFile];
  if((unsigned)[economyToggleProbs getCount] != economyBSSize) {
    if(quitFlag) {
      fprintf(stderr, "%s -- Economy toggle probability file %s number of "
	      "elements (%d) is not equal to economy bit string size (%u)\n",
	      sel_get_name(_cmd), economyToggleProbFile,
	      [economyToggleProbs getCount], economyBSSize);
      exit(1);
    }
    else {
      economyBSSize = (unsigned)[economyToggleProbs getCount];
    }
  }
}

/*

finaliseClimateEconomyToggleProbs

At the end of this method, the climate and economy toggle probability arrays
must be ready for use in the model.

*/

-(void)finaliseClimateEconomyToggleProbs {
  if(climEconCreated) {
    return;
  }
  [self createClimateEconomyToggleProbs];
  [self loadClimateEconomyToggleProbs: YES];
}


@end
